// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file LeafData.h */

#ifndef AMDIS_LEAFDATA_H
#define AMDIS_LEAFDATA_H

#include <list>
#include <map>
#include "MemoryManager.h"
#include "Serializable.h"
#include "FixVec.h"
#include "ElementData.h"
#include "Boundary.h"


namespace AMDiS {

  class LeafDataEstimatableInterface
  {
  public:
    virtual ~LeafDataEstimatableInterface() {};
    virtual void setErrorEstimate(int, double) = 0;
    virtual double getErrorEstimate(int) = 0;
  };

  class LeafDataEstimatable : public ElementData,
			      public LeafDataEstimatableInterface
  {
  public:
    MEMORY_MANAGED(LeafDataEstimatable);

    inline bool isOfType(int type) const {
      if(type == ESTIMATABLE) 
	return true;
      return false;
    };

    class Creator : public CreatorInterface<ElementData>
    {
    public:
      ElementData* create() {
	return NEW LeafDataEstimatable;
      };
    };

    /** \brief
     * constructor
     */
    LeafDataEstimatable(ElementData *decorated = NULL)
      : ElementData(decorated), errorEstimate(0.)
    {};

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    bool refineElementData(Element* parent, 
			   Element* child1,
			   Element* child2,
			   int elType);

    /** \brief
     * 
     */
    void coarsenElementData(Element* parent, 
			    Element* thisChild,
			    Element* otherChild,
			    int elTypeParent);

    /** \brief
     * Sets \ref errorEstimate
     */
    inline void setErrorEstimate(int, double est) { errorEstimate = est; };

    /** \brief
     * Returns \ref errorEstimate
     */
    inline double getErrorEstimate(int) { return errorEstimate; };

    /** \brief
     * Implements ElementData::clone().
     */
    virtual ElementData *clone() const {
      // create new estimatable leaf data
      LeafDataEstimatable *newObj = new LeafDataEstimatable(NULL);

      newObj->errorEstimate = errorEstimate;

      // clone decorated element data (=> deep copy)
      newObj->decorated_ = ElementData::clone();

      // return the clone
      return newObj;
    };

    /** \brief
     * Returns the name of element data type.
     */
    inline std::string getTypeName() const { return "LeafDataEstimatable"; };

    inline const int getTypeID() const { return ESTIMATABLE; };

    void serialize(std::ostream& out) 
    {
      ElementData::serialize(out);
      out.write(reinterpret_cast<const char*>(&errorEstimate), sizeof(double));
    };

    void deserialize(std::istream& in) 
    {
      ElementData::deserialize(in);
      in.read(reinterpret_cast<char*>(&errorEstimate), sizeof(double));
    };

  private:
    double errorEstimate;
  };

  class LeafDataEstimatableVec : public ElementData,
				 public LeafDataEstimatableInterface
  {
  public:
    MEMORY_MANAGED(LeafDataEstimatableVec);

    class Creator : public CreatorInterface<ElementData>
    {
    public:
      ElementData* create() {
	return NEW LeafDataEstimatableVec;
      };
    };

    inline bool isOfType(int type) const {
      if(type == ESTIMATABLE) 
	return true;
      return false;
    };

    /** \brief
     * constructor
     */
    LeafDataEstimatableVec(ElementData *decorated = NULL)
      : ElementData(decorated)
    {
      //errorEstimate.set(0.0);
    };

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    bool refineElementData(Element* parent, 
			   Element* child1,
			   Element* child2,
			   int elType);

    /** \brief
     * 
     */
    void coarsenElementData(Element* parent, 
			    Element* thisChild,
			    Element* otherChild,
			    int elTypeParent);

    /** \brief
     * Sets \ref errorEstimate
     */
    inline void setErrorEstimate(int index, double est) { 
      errorEstimate[index] = est; 
    };

    /** \brief
     * Returns \ref errorEstimate
     */
    inline double getErrorEstimate(int index) { 
      return errorEstimate[index];
    };

    /** \brief
     * Implements ElementData::clone().
     */
    virtual ElementData *clone() const {
      // create new estimatable leaf data
      LeafDataEstimatableVec *newObj = 
	new LeafDataEstimatableVec(NULL);

      newObj->errorEstimate = errorEstimate;

      // clone decorated element data (=> deep copy)
      newObj->decorated_ = ElementData::clone();

      // return the clone
      return newObj;
    };

    void serialize(std::ostream& out) 
    {
      ElementData::serialize(out);
      unsigned int size = errorEstimate.size();
      out.write(reinterpret_cast<const char*>(&size), sizeof(unsigned int));
      std::map<int, double>::iterator it;
      for(it = errorEstimate.begin(); it != errorEstimate.end(); ++it) {
	out.write(reinterpret_cast<const char*>(&(it->first)), sizeof(int));
	out.write(reinterpret_cast<const char*>(&(it->second)), sizeof(double));
      }
    };

    void deserialize(std::istream& in) 
    {
      ElementData::deserialize(in);
      unsigned int i, size;
      in.read(reinterpret_cast<char*>(&size), sizeof(unsigned int));
      for(i = 0; i < size; i++) {
	int index;
	double estimate;
	in.read(reinterpret_cast<char*>(&index), sizeof(int));
	in.read(reinterpret_cast<char*>(&estimate), sizeof(double));
	errorEstimate[index] = estimate;
      }
    };

    std::string getTypeName() const
    {
      return "LeafDataEstimatableVec";
    };
  
    inline const int getTypeID() const { return ESTIMATABLE; };

  private:
    std::map<int, double> errorEstimate;
  };

  class LeafDataCoarsenableInterface
  {
  public:
    virtual ~LeafDataCoarsenableInterface() {};

    /** \brief
     * Sets \ref coarseningError
     */
    virtual void setCoarseningErrorEstimate(int index, double est)=0;

    /** \brief
     * Returns \ref coarseningError
     */
    virtual double getCoarseningErrorEstimate(int index) = 0;
  };

  class LeafDataCoarsenable : public ElementData,
			      public LeafDataCoarsenableInterface
  {
  public:
    MEMORY_MANAGED(LeafDataCoarsenable);

    class Creator : public CreatorInterface<ElementData>
    {
    public:
      ElementData* create() {
	return NEW LeafDataCoarsenable;
      };
    };

    inline bool isOfType(int type) const {
      if(type == COARSENABLE) 
	return true;
      return false;
    };

    /** \brief
     * constructor
     */
    LeafDataCoarsenable(ElementData *decorated = NULL)
      : ElementData(decorated), coarseningError(0.0)
    {};

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    bool refineElementData(Element* parent, 
			   Element* child1,
			   Element* child2,
			   int elType);

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    void coarsenElementData(Element* parent, 
			    Element* thisChild,
			    Element* otherChild,
			    int elTypeParent);

    /** \brief
     * Implements ElementData::clone().
     */
    inline ElementData *clone() const {
      // create new estimatable leaf data
      LeafDataCoarsenable *newObj = new LeafDataCoarsenable(NULL);

      // clone decorated element data (=> deep copy)
      newObj->decorated_ = ElementData::clone();

      // return the clone
      return newObj;
    };

    /** \brief
     * Sets \ref coarseningError
     */
    virtual void setCoarseningErrorEstimate(int , double est) { 
      coarseningError = est; 
    };

    /** \brief
     * Returns \ref coarseningError
     */
    virtual double getCoarseningErrorEstimate(int) { 
      return coarseningError; 
    };

    // ===== Serializable implementation =====

    void serialize(std::ostream& out) 
    {
      ElementData::serialize(out);
      out.write(reinterpret_cast<const char*>(&coarseningError), sizeof(double));
    };

    void deserialize(std::istream& in) 
    {
      ElementData::deserialize(in);
      in.read(reinterpret_cast<char*>(&coarseningError), sizeof(double));
    };

    std::string getTypeName() const
    {
      return "LeafDataCoarsenable";
    };
  
    inline const int getTypeID() const { return COARSENABLE; };

  private:
    double coarseningError; /**< \brief coarsening error */
  };



  class LeafDataCoarsenableVec : public ElementData,
				 public LeafDataCoarsenableInterface
  {
  public:
    MEMORY_MANAGED(LeafDataCoarsenableVec);

    class Creator : public CreatorInterface<ElementData>
    {
    public:
      ElementData* create() {
	return NEW LeafDataCoarsenableVec;
      };
    };

    inline bool isOfType(int type) const {
      if(type == COARSENABLE) 
	return true;
      return false;
    };

    /** \brief
     * constructor
     */
    LeafDataCoarsenableVec(ElementData *decorated = NULL)
      : ElementData(decorated) 
    {};

    /** \brief
     * Implements ElementData::clone().
     */
    inline ElementData *clone() const {
      // create new estimatable leaf data
      LeafDataCoarsenableVec *newObj = 
	new LeafDataCoarsenableVec(NULL);

      newObj->coarseningError = coarseningError; 

      // clone decorated leaf data (=> deep copy)
      newObj->decorated_ = ElementData::clone();

      // return the clone
      return newObj;
    };

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    bool refineElementData(Element* parent, 
			   Element* child1,
			   Element* child2,
			   int elType);

    /** \brief
     * Refinement of parent to child1 and child2.
     */
    void coarsenElementData(Element* parent, 
			    Element* thisChild,
			    Element* otherChild,
			    int elTypeParent);

    /** \brief
     * Sets \ref coarseningError
     */
    virtual void setCoarseningErrorEstimate(int index, double est) { 
      coarseningError[index] = est; 
    };

    /** \brief
     * Returns \ref coarseningError
     */
    virtual double getCoarseningErrorEstimate(int index) { 
      return coarseningError[index]; 
    };

    // ===== Serializable implementation =====

    void serialize(std::ostream& out) 
    {
      ElementData::serialize(out);
      unsigned int size = coarseningError.size();
      out.write(reinterpret_cast<const char*>(&size), sizeof(unsigned int));
      std::map<int, double>::iterator it;
      for(it = coarseningError.begin(); it != coarseningError.end(); ++it) {
	out.write(reinterpret_cast<const char*>(&(it->first)), sizeof(int));
	out.write(reinterpret_cast<const char*>(&(it->second)), sizeof(double));
      }
    };

    void deserialize(std::istream& in) 
    {
      ElementData::deserialize(in);
      unsigned int i, size;
      in.read(reinterpret_cast<char*>(&size), sizeof(unsigned int));
      for(i = 0; i < size; i++) {
	int index;
	double estimate;
	in.read(reinterpret_cast<char*>(&index), sizeof(int));
	in.read(reinterpret_cast<char*>(&estimate), sizeof(double));
	coarseningError[index] = estimate;
      }
    };

    std::string getTypeName() const 
    {
      return "LeafDataCoarsenableVec";
    };
  
    inline const int getTypeID() const { return COARSENABLE; };

  private:
    std::map<int, double> coarseningError; /**< \brief coarsening error */
  };



  class LeafDataPeriodic : public ElementData
  {
  public:
    MEMORY_MANAGED(LeafDataPeriodic);

    class Creator : public CreatorInterface<ElementData>
    {
    public:
      ElementData* create() {
	return NEW LeafDataPeriodic;
      };
    };

    inline bool isOfType(int type) const {
      if(type == PERIODIC) 
	return true;
      return false;
    };


    class PeriodicInfo : public Serializable
    {
    public:
      PeriodicInfo() 
	:	periodicCoords(NULL)
      {};

      PeriodicInfo(int mode,
		   BoundaryType t,
		   int side,
		   const DimVec<WorldVector<double> > *coords);

      virtual ~PeriodicInfo() {
	if(periodicCoords) DELETE periodicCoords;
      };

      PeriodicInfo(const PeriodicInfo &rhs);

      void serialize(std::ostream &out) {
	out.write(reinterpret_cast<const char*>(&periodicMode), sizeof(int));
	out.write(reinterpret_cast<const char*>(&type), sizeof(BoundaryType));
	out.write(reinterpret_cast<const char*>(&elementSide), sizeof(int));
	int i, size;
	if(periodicCoords) {
	  size = periodicCoords->getSize();
	  out.write(reinterpret_cast<const char*>(&size), sizeof(int));
	  for(i = 0; i < size; i++) {
	    (*periodicCoords)[i].serialize(out);
	  }
	} else {
	  size = 0;
	  out.write(reinterpret_cast<const char*>(&size), sizeof(int));
	}
      };

      void deserialize(std::istream &in)
      {
	in.read(reinterpret_cast<char*>(&periodicMode), sizeof(int));
	in.read(reinterpret_cast<char*>(&type), sizeof(BoundaryType));
	in.read(reinterpret_cast<char*>(&elementSide), sizeof(int));

	int i, size;
	in.read(reinterpret_cast<char*>(&size), sizeof(int));
	if(periodicCoords) DELETE periodicCoords;
	if(size == 0) {
	  periodicCoords = NULL;
	} else {
	  periodicCoords = NEW DimVec<WorldVector<double> >(size-1, NO_INIT);
	  for(i = 0; i < size; i++) {
	    (*periodicCoords)[i].deserialize(in);
	  }
	}
      };

      int periodicMode;
      BoundaryType type;
      int elementSide;
      DimVec<WorldVector<double> > *periodicCoords;
    };

  public:
    /** \brief
     * constructor
     */
    LeafDataPeriodic(ElementData *decorated = NULL)
      : ElementData(decorated)
		    //newDOF(-1)
    {};

    /** \brief
     * Destructor.
     */
    ~LeafDataPeriodic() {};

    /** \brief
     * Implements LeafData::clone().
     */
    inline ElementData *clone() const {
      LeafDataPeriodic *newObj = new LeafDataPeriodic;
      newObj->decorated_ = ElementData::clone();
      return newObj;
    };

    inline void addPeriodicInfo(int mode,
				BoundaryType type,
				int side,
				const DimVec<WorldVector<double> > *coords)
    {
      PeriodicInfo periodicInfo(mode, type, side, coords);
      periodicInfoList.push_back(periodicInfo);
    };

    inline std::list<PeriodicInfo>& getInfoList() {
      return periodicInfoList;
    };

    // ===== Serializable implementation =====

    void serialize(std::ostream& out) 
    {
      ElementData::serialize(out);
      //out.write(reinterpret_cast<const char*>(&newDOF), sizeof(DegreeOfFreedom));
      unsigned int size = periodicInfoList.size();
      out.write(reinterpret_cast<const char*>(&size), sizeof(unsigned int));
      std::list<PeriodicInfo>::iterator it;
      for(it = periodicInfoList.begin(); it != periodicInfoList.end(); ++it) {
	it->serialize(out);
      }
    };

    void deserialize(std::istream& in) 
    {
      ElementData::deserialize(in);
      //in.read(reinterpret_cast<char*>(&newDOF), sizeof(DegreeOfFreedom));
      unsigned int size;
      in.read(reinterpret_cast<char*>(&size), sizeof(unsigned int));
      periodicInfoList.resize(size);
      std::list<PeriodicInfo>::iterator it;
      for(it = periodicInfoList.begin(); it != periodicInfoList.end(); ++it) {
	it->deserialize(in);
      }
    };

    std::string getTypeName() const
    {
      return "LeafDataPeriodic";
    };
  
    inline const int getTypeID() const { return PERIODIC; };

    bool refineElementData(Element* parent, 
			   Element* child1,
			   Element* child2,
			   int elType);

    //  inline void setNewDOF(DegreeOfFreedom nd) { newDOF = nd; };

  private:
    //DegreeOfFreedom newDOF;
    std::list<PeriodicInfo> periodicInfoList;

    friend class LeafDataPeriodicRefinable;
    friend class LeafDataPeriodicCoarsenable;
  };

}

#endif  // AMDIS_LEAFDATA_H

