#include "Preconditioner.h"

namespace AMDiS {

  template<typename VectorType>
  OResSolver<VectorType>::OResSolver(std::string name)
    : OEMSolver<VectorType>(name),
      Ad(NULL), d(NULL), d_old(NULL), r(NULL), h(NULL), Ad_old(NULL), Ah(NULL)
  {
  }

  template<typename VectorType>
  OResSolver<VectorType>::~OResSolver()
  {}

  template<typename VectorType>
  void OResSolver<VectorType>::init()
  {
    Ad     = this->vectorCreator->create();
    d      = this->vectorCreator->create();
    d_old  = this->vectorCreator->create();
    r      = this->vectorCreator->create();
    h      = this->vectorCreator->create();
    Ad_old = this->vectorCreator->create();
    Ah     = this->vectorCreator->create();
  }

  template<typename T>
  void OResSolver<T>::exit()
  {
    this->vectorCreator->free(Ad);
    this->vectorCreator->free(d);
    this->vectorCreator->free(d_old);
    this->vectorCreator->free(r);
    this->vectorCreator->free(h);
    this->vectorCreator->free(Ad_old);
    this->vectorCreator->free(Ah);
  }

  template<typename VectorType>
  int OResSolver<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
					  VectorType *x, VectorType *b)
  {
    FUNCNAME("OResSolver::solveSystem");
  
    double alpha, gamma, sigma;
    double daad, daad_old, rad, res, old_res = -1.0;
    int   iter; //, dim2;
    VectorType *swap = NULL;

    /*--------------------------------------------------------------------------*/
    /*---  Initalization  ------------------------------------------------------*/
    /*--------------------------------------------------------------------------*/

    // r = Ax
    matVec->matVec(NoTranspose, *x, *r);

    // r = - b + r
    *r -= *b;
  
    // h = r
    *h = *r;

    if (this->leftPrecon) this->leftPrecon->precon(h);

    // d = h
    *d = *h;

    /*--- d_old = 0 and Ad_old = 0 ---------------*/
    set(*d_old, 0.0);
    set(*Ad_old, 0.0);

    // Ad = A*d
    matVec->matVec(NoTranspose, *d, *Ad);

    daad_old = 1.0;

    /*--------------------------------------------------------------------------*/
    /*---  check initial residual  ---------------------------------------------*/
    /*--------------------------------------------------------------------------*/
  
    res = norm(r);

    START_INFO();
    if (SOLVE_INFO(0, res, &old_res))
      return(0);

    for (iter = 1; iter <= this->max_iter; iter++) {
      /*--- compute r.Ad and d.A.A.d  ---------------------------------------------*/
      // rad = r * Ad
      rad = *r * *Ad;
      
      // daad = Ad * Ad
      daad = *Ad * *Ad;
      
      if (daad <= 1.e-30) {
	BREAK_INFO("(Ad,d)_2 = 0", iter, res, &old_res);
	return(iter);
      }

      /*--- update u and r  ------------------------------------------------------*/
      
      alpha= rad/daad;
      
      // x = -alpha d + x
      axpy(-alpha, *d, *x);
      
      // r = - alpha Ad + r
      axpy(-alpha, *Ad, *r);
      
      res = norm(r);
      
      if (SOLVE_INFO(iter, res, &old_res))
	return(iter);
      
      // h = Ad
      *h = *Ad;
      
      if (this->leftPrecon) this->leftPrecon->precon(h);
      
      // Ah = A * h
      matVec->matVec(NoTranspose, *h, *Ah);
      
      // gamma = Ah * Ad / daad
      gamma = (*Ah * *Ad)/daad;
      sigma = daad/daad_old;
      
      /*--- compute new d.Ad, save old values!  ----------------------------------*/
      // swap d and d_old
      swap = d_old;
      d_old = d;
      d = swap;
      
      // swap Ad_old and Ad
      swap = Ad_old;
      Ad_old = Ad;
      Ad = swap;
      
      /*--- d,Ad *= -sigma --------------*/
      *d  *= -sigma;
      *Ad *= -sigma;
	
      /*--- d,Ad -= gamma*d_old,Ad_old --*/
      axpy(-gamma, *d_old, *d);
      axpy(-gamma, *Ad_old, *Ad);
      
      // d = h + d
      *d += *h;
      
      // Ad = Ah + Ad
      *Ad += *Ah;
      
      daad_old = daad;
    }

    return 0; // never reached !
  }

}
