#include "TFQMR.h"
#include "Preconditioner.h"

namespace AMDiS {

  template<typename VectorType>
  TFQMR<VectorType>::TFQMR(std::string name)
    : OEMSolver<VectorType>(name),
      TOL_(1.e-25)
  {
    FUNCNAME("TFQMR::TFQMR()");
  }

  template<typename VectorType>
  TFQMR<VectorType>::~TFQMR()
  {}

  template<typename VectorType>
  void TFQMR<VectorType>::init()
  {
    r0_ = this->vectorCreator->create();
    w_ = this->vectorCreator->create();
    y1_ = this->vectorCreator->create();
    y2_ = this->vectorCreator->create();
    v_ = this->vectorCreator->create();
    d_ = this->vectorCreator->create();
    tmp_ = this->vectorCreator->create();
  }

  template<typename VectorType>
  void TFQMR<VectorType>::exit()
  {
    if (r0_) {
      this->vectorCreator->free(r0_);
      this->vectorCreator->free(w_);
      this->vectorCreator->free(y1_);
      this->vectorCreator->free(y2_);
      this->vectorCreator->free(v_);
      this->vectorCreator->free(d_);
      this->vectorCreator->free(tmp_);
    }
  }

  template<typename VectorType>
  int TFQMR<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
					    VectorType *x, 
					    VectorType *b)
  {
    DOFMatrix *m = (dynamic_cast<StandardMatVec<DOFMatrix, VectorType> *>(matVec))->getMatrix();

    matVec->matVec(NoTranspose, *x, *r0_);
    xpay(-1.0, *b, *r0_);
    double tau = norm(r0_);
    std::cout << "NORM: " <<  tau << std::endl;

    for (int iter = 0; iter < 1000; iter++) {
      for (int i = 0; i < x->getSize(); i++) {
	(*y1_)[i] = (*b)[i];

	for (int j = 0; j < static_cast<int>((*m)[i].size()); j++) {
	  MatEntry a = (*m)[i][j];

	  if (a.col != i) {
	    (*y1_)[i] -= (*x)[i] * a.entry;
	  }
	}

	(*y1_)[i] /= (*m)[i][0].entry;
      }

      *x = *y1_;
    }

    matVec->matVec(NoTranspose, *x, *r0_);
    xpay(-1.0, *b, *r0_);
    tau = norm(r0_);
    std::cout << "NORM: " <<  tau << std::endl;

    return 1000;
  }

  //  template<typename VectorType>
//   int TFQMR<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
// 					    VectorType *x, 
// 					    VectorType *b)
//   {
//     FUNCNAME("TFQMR::solveSystem()");    

//     double old_res = -1.0;

//     // If norm of b is smaller than the tolarance, we can assume b to be zero.
//     // Hence, x = 0 is the solution of the linear system.
//     if (norm(b) < TOL_) {
//       INFO(this->info, 2)("b == 0, x = 0 is the solution of the linear system\n");
//       set(*x, 0.0);
//       this->residual = 0.0;
//       return(0);
//     }

//     // r_0 = b - Ax, where r_0 is already stored as the first vector in the
//     // matrix V.
//     matVec->matVec(NoTranspose, *x, *r0_);
//     xpay(-1.0, *b, *r0_);

//     if (this->leftPrecon) 
//       this->leftPrecon->precon(r0_);

//     *y1_ = *r0_;
//     *w_ = *r0_;
//     double tau = norm(r0_);

//     if (tau < this->tolerance) {
//       this->residual = tau;
//       return(0);
//     }

//     matVec->matVec(NoTranspose, *y1_, *v_);
//     d_->set(0.0);
//     double eta = 0.0;
//     double theta = 0.0;
//     double rho1 = tau * tau;
//     double rho2 = 0.0;
//     double alpha = 0.0;
//     double beta = 0.0;
//     double c = 0.0;

//     START_INFO();
//     for (int j = 1; j <= this->max_iter; j++) {      
//       rho2 = *v_ * *r0_;
//       alpha = rho1 / rho2;
//       *y2_ = *y1_;
//       axpy(-alpha, *v_, *y2_);

//       for (int m = 2 * j - 1; m <= 2 * j; m++) {
// 	matVec->matVec(NoTranspose, *y1_, *tmp_);
// 	axpy(-alpha, *tmp_, *w_);

// 	if (this->leftPrecon) 
// 	  this->leftPrecon->precon(w_);

// 	xpay(theta * theta * eta / alpha, *y1_, *d_);


// 	theta = norm(w_) / tau;
// 	c = 1 / sqrt(theta * theta + 1);
// 	eta = c * c * alpha;
// 	axpy(eta, *d_, *x);
// 	tau = tau * theta * c;
// 	this->residual = sqrt(m + 1) * tau;	
//         *y1_ = *y2_;
//       }

//       if (SOLVE_INFO(j - 1, this->residual, &old_res))
// 	return(j - 1); 

//       rho2 = *w_ * *r0_;
//       beta = rho2 / rho1;
//       *y1_ = *w_;
//       axpy(beta, *y2_, *y1_);

      
//       matVec->matVec(NoTranspose, *y2_, *tmp_);

//       if (this->leftPrecon) 
// 	this->leftPrecon->precon(tmp_);

//       xpay(beta, *tmp_, *v_);

//       matVec->matVec(NoTranspose, *y1_, *tmp_);

//       if (this->leftPrecon) 
// 	this->leftPrecon->precon(tmp_);


//       xpay(beta, *tmp_, *v_);

//       rho1 = rho2;
//     }

//     return 0;
//   }

}
