#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

class Project : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(Project);

  /** \brief
   * Constructor
   */
  Project(int i) : comp(i) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const
  {
    static double result;
    result = x[comp];
    return result;  
  };

protected:
  /** \brief
   * coordinate plane to project at
   */
  int comp;
};


/** \brief
 * RHS function
 */
class ConstantFct : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(ConstantFct);

  ConstantFct(double c) 
    : AbstractFunction<double, WorldVector<double> >(0) ,
      constant(c)
  {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const 
  {
    static double result = 0.0;
    result = constant;
    return result;
  };

protected:
  double constant;
};

class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  F() 
    : AbstractFunction<double, WorldVector<double> >(0)
  {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const 
  {
    static double result = 0.0;
    result = (x[0] == 0.0) || (x[0] == 1.0) ? 0.0 : 1.0;
    return result;
  };

protected:
  double constant;
};

// ===========================================================================
// ===== Updater =============================================================
// ===========================================================================

/** \brief
 * Updater for the Stokes operator
 */

class StokesUpdater: public  NonLinUpdaterVec 
{
public:
  /** \brief
   * Constructor.
   */
  StokesUpdater(Matrix<DOFMatrix*> *m, SystemVector *sol) : 
    NonLinUpdaterVec(m),sol_(sol) 
  {};


  virtual void update(bool updateDF, SystemVector *F) {
 
    FUNCNAME("StokesUpdater::update()");

    int i, j, size = df->getNumRows();
    Matrix<DOFMatrix*>* lhsMatrix=NULL;
    static Matrix<DOFMatrix*>* lhsMat=NULL; // Storage for lhs matrix if DF is not updated
    
    TraverseStack stack;
    ElInfo    *el_info;
    Flag     fill_flag;
    
    TEST_EXIT(F || df)("neither F nor df are set\n");
    
    if ((!updateDF) && (F==NULL)) {
      WARNING("No RHS vector and no update for system matrix! Updater does nothing!\n");
      return;
    }
    
    const FiniteElemSpace *feSpace = 
      F ? 
      F->getDOFVector(0)->getFESpace() : 
      (*df)[0][0]->getFESpace();
    
    if (updateDF) { 
      TEST_EXIT(df)("df not set but update tried!\n");
      for(i = 0; i < size; i++) {
	for(j = 0; j < size; j++) {
	  if((*df)[i][j]) {
	    (*df)[i][j]->clear();
	  }
	}
      }
      lhsMatrix=df; // lhs matrix is DF, which has to be updated.
    }
    else { // DF must not be changed, so we provide storage for lhs matrix.
      if (lhsMat==NULL) {
	lhsMat=NEW Matrix<DOFMatrix*>(size,size);
	for(i = 0; i < size; i++) {
	  for(j = 0; j < size; j++) {
	    if ((!df)||(*df)[i][j]) {
	      (*lhsMat)[i][j]=NEW DOFMatrix(*((*df)[i][j]));
	      (*lhsMat)[i][j]->clear();
	    }
	    else {
	      (*lhsMat)[i][j]=NULL;
	    }
	  }
	}
      }
      lhsMatrix=lhsMat;
    }
    
    BasisFunction *basFcts = const_cast<BasisFunction*>(feSpace->getBasisFcts());
    
    if (F) { // clear RHS vector if neccessary.
      for(i = 0; i < size; i++) {
	F->getDOFVector(i)->set(0.0);
      }
    }

   fill_flag = 
      Mesh::CALL_LEAF_EL|
      Mesh::FILL_COORDS|
      Mesh::FILL_BOUND|
      Mesh::FILL_DET|
      Mesh::FILL_GRD_LAMBDA;
    
    el_info = stack.traverseFirst(feSpace->getMesh(), -1, fill_flag);
    while (el_info) {
      const BoundaryType *bound = basFcts->getBound(el_info, NULL);
      // assemble LHS matrix.
      for(i = 0; i < size; i++) {
	for(j = 0; j < size; j++) {
	  if((*lhsMatrix)[i][j]) {
	    (*lhsMatrix)[i][j]->assemble(1.0, el_info, bound);
	    if((*lhsMatrix)[i][j]->getBoundaryManager())
	      (*lhsMatrix)[i][j]->getBoundaryManager()->fillLocalBCs(el_info,
							(*lhsMatrix)[i][j]);
	  }
	}
      }
      
      
      if (F) { // assemble RHS vector if neccessary
	for(i = 0; i < size; i++) {
	  F->getDOFVector(i)->assemble(1.0, el_info, bound);
	  if(F->getDOFVector(i)->getBoundaryManager())
	    F->getDOFVector(i)->getBoundaryManager()->fillLocalBCs(el_info, 
							F->getDOFVector(i));
	}
      }
      
      el_info = stack.traverseNext(el_info);
    }
    
    if (F) { 
      SystemVector lonR(*F);

      // final RHS vector is LHS matrix times old solution minus assembled RHS vector 
      // i.e. the residual of the linear equation.
    
      StandardMatVec<Matrix<DOFMatrix*>,SystemVector> matVec(lhsMatrix);

      matVec.matVec(NoTranspose,*sol_,lonR);
      *F *= -1.0;
      *F +=lonR;

    }
  };

private:
  /** \brief
   * Solution of the nonlinear Problem. Needed to calculate RHS contribution of Matrix.
   */
  SystemVector* sol_;

};

// ===========================================================================
// ===== instationary problem ================================================
// ===========================================================================

/** \brief
 * Instationary problem
 */
class Navierstokes : public ProblemInstatVec
{
public:
  MEMORY_MANAGED(Navierstokes);

  /** \brief
   *  Constructor
   */
  Navierstokes(ProblemVec *navierstokesSpace) 
    : ProblemInstatVec("navierstokes", navierstokesSpace)
  {
    int i, dow = Global::getGeo(WORLD);
    TEST_EXIT(dow == navierstokesSpace->getNumComponents() - 1)
      ("number of components != dow + 1\n");
    boundaryFct.resize(dow+1, NULL);

    epsilon = -1.0;
    GET_PARAMETER(0, "epsilon", "%f", &epsilon);
    TEST_EXIT(epsilon != -1.0)("invalid epsilon\n");
  };

  // ===== initial problem methods =====================================

  /** \brief
   * Used by \ref problemInitial to solve the system of the initial problem
   */
  void solve(AdaptInfo *adaptInfo) 
  {
    int i, size = problemStat->getNumComponents()-1;
    for(i = 0; i < size; i++) {
      //problemStat->getMesh(i)->dofCompress();
      problemStat->getSolution()->getDOFVector(i)->interpol(boundaryFct[i]);
    }
  };

  /** \brief
   * Used by \ref problemInitial to do error estimation for the initial
   * problem.
   */
  void estimate(AdaptInfo *adaptInfo) 
  {
    int i, size = problemStat->getNumComponents()-1;
    double errMax, errSum;

    for(i = 0; i < size; i++) {
      errSum = 
	Error<double>::L2Err(*boundaryFct[i],
			     *(problemStat->getSolution()->getDOFVector(i)), 
			     0, &errMax, false);

      adaptInfo->setEstSum(errSum, i);
    }
  };

  /** \brief
   * Sets \ref boundaryFct;
   */
  void setBoundaryFct(AbstractFunction<double, WorldVector<double> > *fct, int i) 
  {
    boundaryFct[i] = fct;
  };

  /** \brief
   * Set the time in all needed functions! Called by adaption loop.
   */
  void setTime(AdaptInfo *adaptInfo) 
  {
    time = adaptInfo->getTime();
    tau1 = 1.0 / adaptInfo->getTimestep();
  };
  
  /** \brief
   * Returns address of \ref time.
   */
  double *getTimePtr() {
    return &time;
  };

  /** \brief
   * Returns address of \ref tau1.
   */
  double *getTau1Ptr() {
    return &tau1;
  };

  /** \brief
   * Implementation of ProblemInstatBase::closeTimestep()
   * pressure correction step
   */
  void closeTimestep(AdaptInfo *adaptInfo) {
    problemStat->writeFiles(adaptInfo, true);

    WAIT;

    int i, numComponents = problemStat->getNumComponents();
    int dow = Global::getGeo(WORLD);

    double timestep = adaptInfo->getTimestep();

    static DOFVector<double> pInterpol(problemStat->getFESpace(0), "p interpolated");
    static DOFVector<WorldVector<double> > gradPInterpol(problemStat->getFESpace(0), "gradient p interpolated");
    DOFVector<double> *p = problemStat->getSolution()->getDOFVector(numComponents-1);

    pInterpol.interpol(p, timestep * epsilon);
    pInterpol.getRecoveryGradient(&gradPInterpol);

    for(i = 0; i < dow; i++) {
      DOFVector<double>::Iterator 
	vIt(problemStat->getSolution()->getDOFVector(i), USED_DOFS);
      DOFVector<WorldVector<double> >::Iterator 
	gradIt(&gradPInterpol, USED_DOFS);

      for(vIt.reset(), gradIt.reset(); !vIt.end(); ++vIt, ++gradIt) {
	*vIt -= (*gradIt)[i];
      }
    }

    problemStat->writeFiles(adaptInfo, true);
  };

private:
  /** \brief
   * boundary function
   */
  std::vector<AbstractFunction<double, WorldVector<double> >*> boundaryFct;

  /** \brief
   * time
   */
  double time;

  /** \brief
   * 1.0/timestep
   */
  double tau1;

  /** \brief
   */
  double epsilon;
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char** argv)
{
  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc == 2)("usage: navierstokes initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // read nu (1/Re)
  double nu = -1.0;
  GET_PARAMETER(0, "nu", "%f", &nu);
  TEST_EXIT(nu != -1.0)("invalid nu\n");

  // ===== create and init stationary problem =====
  ProblemNonLinVec *navierstokesSpace = NEW ProblemNonLinVec("navierstokes->space");
  navierstokesSpace->initialize(INIT_ALL&~INIT_UPDATER&~INIT_NONLIN_SOLVER);
  StokesUpdater* stokesUpdater=new StokesUpdater(navierstokesSpace->getSystemMatrix(),
 						 navierstokesSpace->getSolution());
  navierstokesSpace->setUpdater(stokesUpdater);

  //navierstokesSpace->setUpdater(NEW NonLinUpdaterVec(navierstokesSpace->getSystemMatrix()));
  navierstokesSpace->initialize(INIT_NONLIN_SOLVER);

  int numComponents = navierstokesSpace->getNumComponents();

  // ===== create instationary problem =====
  Navierstokes *navierstokes = NEW Navierstokes(navierstokesSpace);
  navierstokes->initialize(INIT_ALL);

  AdaptInfo *adaptInfo = NEW AdaptInfo("navierstokes->adapt",
				       numComponents);

  // create initial adapt
  AdaptInfo *adaptInfoInitial = NEW AdaptInfo("navierstokes->initial->adapt",
					      numComponents);

  // create instationary adapt
  AdaptInstationary *adaptInstat =
    NEW AdaptInstationary("navierstokes->adapt",
			  navierstokesSpace,
			  adaptInfo,
			  navierstokes,
			  adaptInfoInitial);

  // ===== create boundary functions =====
  
  double *tau1 = navierstokes->getTau1Ptr();
  int i;

  // === boundary functions ===
  ConstantFct *boundaryFct = NEW ConstantFct(0.0);

  for(i = 0; i < numComponents-1; i++) {
    navierstokes->setBoundaryFct(boundaryFct, i);
    navierstokesSpace->addDirichletBC(1, i, boundaryFct);
  }
  navierstokesSpace->addDirichletBC(2, 0, NEW F());
  navierstokesSpace->addDirichletBC(2, 1, boundaryFct);
  navierstokes->setBoundaryFct(boundaryFct, numComponents-1);
  

  // ===== create operators =====

  // ===== velocity =====
  for(i = 0; i < numComponents-1; i++) {
    // create laplace
    Operator *laplaceV = NEW Operator(Operator::MATRIX_OPERATOR,
				      navierstokesSpace->getFESpace(i),
				      navierstokesSpace->getFESpace(i));

    laplaceV->addSecondOrderTerm(NEW FactorLaplace_SOT(nu));
    navierstokesSpace->addMatrixOperator(laplaceV, i, i);

    // create zero order operator
    Operator *dtv = NEW Operator(Operator::MATRIX_OPERATOR |
				 Operator::VECTOR_OPERATOR,
				 navierstokesSpace->getFESpace(i),
				 navierstokesSpace->getFESpace(i));

    dtv->setUhOld(navierstokes->getOldSolution()->getDOFVector(i));
    dtv->addZeroOrderTerm(NEW Simple_ZOT);

    navierstokesSpace->addMatrixOperator(dtv, i, i, tau1, tau1);
    navierstokesSpace->addVectorOperator(dtv, i, tau1, tau1);

    Operator *gradP = NEW Operator(Operator::VECTOR_OPERATOR,
				   navierstokesSpace->getFESpace(i));
    gradP->addZeroOrderTerm(NEW FctGradient_ZOT(navierstokes->getOldSolution()->
						getDOFVector(numComponents-1),
						NEW Project(i)));
    double minusOne = 1.0;
    navierstokesSpace->addVectorOperator(gradP, i, &minusOne, &minusOne);
  }

  // ===== pressure =====
  Operator *laplaceP = NEW Operator(Operator::MATRIX_OPERATOR,
				    navierstokesSpace->getFESpace(numComponents-1),
				    navierstokesSpace->getFESpace(numComponents-1));

  laplaceP->addSecondOrderTerm(NEW Laplace_SOT());

  navierstokesSpace->addMatrixOperator(laplaceP, 
				       numComponents-1, 
				       numComponents-1);
  
  WorldVector<double> b;

  for(i = 0; i < numComponents-1; i++) {
    Operator *divV = NEW Operator(Operator::MATRIX_OPERATOR,
				  navierstokesSpace->getFESpace(numComponents-1),
				  navierstokesSpace->getFESpace(i));

    b.set(0.0);
    b[i] = 1.0;

    divV->addFirstOrderTerm(NEW Vector_FOT(b), GRD_PHI);

    navierstokesSpace->addMatrixOperator(divV, numComponents-1, i, tau1, tau1);    
  }

  // ===== start adaption loop =====
  adaptInstat->adapt();
}
